#ifndef LSM_TRIE
#define LSM_TRIE

#include <stdlib.h>

#include "lsm.h"
#include "lsm/bt.h"
#include "lsm/str.h"

/**
 * A struct representing a trie
 */
typedef struct lsm_trie lsm_trie;

/**
 * A node inside an `lsm_trie` trie
 */
typedef struct lsm_trie_node lsm_trie_node;

/**
 * Initialize a new trie.
 *
 * @param ptr where to store the newly allocated pointer
 */
lsm_error lsm_trie_init(lsm_trie **ptr);

/**
 * Deallocate an entire trie, including all its nodes
 *
 * @param trie trie to free
 */
void lsm_trie_free(lsm_trie *trie);

/**
 * Insert a new element into the trie using the specified key.
 *
 * @param trie trie to insert into
 * @param key key to insert data with
 * @param data data to insert
 */
lsm_error lsm_trie_insert(lsm_trie *trie, const lsm_str *key, void *data);

/**
 * Search for an element in the trie.
 *
 * @param out where to store data pointer, if present
 * @param trie trie to search in
 * @param key key to search with
 */
lsm_error lsm_trie_search(void **out, const lsm_trie *trie, const lsm_str *key);

/**
 * Remove an element from the trie.
 *
 * @param out where to store the removed data pointer, if present.
 * @param trie trie to remove from
 * @param key key to remove
 */
lsm_error lsm_trie_remove(void **out, lsm_trie *trie, const lsm_str *key);

/**
 * Return the size of a trie
 *
 * @param trie trie to return size for
 */
uint64_t lsm_trie_size(const lsm_trie *trie);

/**
 * Represents an in-flight iterator over a trie
 */
typedef struct lsm_trie_iterator lsm_trie_iterator;

/**
 * Initialize an iterator to iterate over all entries with keys starting
 * with the given prefix.
 *
 * @param out pointer to store iterator pointer in
 * @param trie trie to iterate
 * @param prefix prefix of the keys; a zero-length string means iterating over
 *  the entire trie; NULL is interpreted as a zero-length string
 */
lsm_error lsm_trie_iter(lsm_trie_iterator **out, const lsm_trie *trie,
                        const lsm_str *prefix);

/**
 * Advance the given iterator.
 *
 * @param out pointer to store next data pointer in
 * @param iter iterator to advance
 * @return `lsm_error_ok` if next element has been returned, `lsm_error_done` if
 *  no more elements are present, or some other error code
 */
lsm_error lsm_trie_iter_next(void **out, lsm_trie_iterator *iter);

/**
 * Free the given iterator.
 */
void lsm_trie_iter_free(lsm_trie_iterator *iter);

#endif
