#include <stdlib.h>

#include "lsm.h"
#include "lsm/trie_internal.h"

lsm_error lsm_trie_node_init(lsm_trie_node **ptr) {
  lsm_trie_node *node = calloc(1, sizeof(lsm_trie_node));

  if (node == NULL) {
    return lsm_error_failed_alloc;
  }

  *ptr = node;

  return lsm_error_ok;
}

void lsm_trie_node_free(lsm_trie_node *node) {
  lsm_bt_clear(&node->bt);
  lsm_str_zero(&node->skip);
  free(node);
}

lsm_error lsm_trie_init(lsm_trie **ptr) {
  lsm_trie *trie = calloc(1, sizeof(lsm_trie));

  if (trie == NULL) {
    return lsm_error_failed_alloc;
  }

  lsm_trie_node *root;
  lsm_error res = lsm_trie_node_init(&root);

  if (res != lsm_error_ok) {
    free(trie);

    return res;
  }

  trie->root = root;
  *ptr = trie;

  return lsm_error_ok;
}

uint64_t lsm_trie_size(const lsm_trie *trie) { return trie->size; }

lsm_error lsm_trie_insert(lsm_trie *trie, const lsm_str *key, void *data) {
  // NULL is not allowed as a data value, as it's used to indicate a lack of
  // data
  if (data == NULL) {
    return lsm_error_null_value;
  }

  uint64_t key_len = lsm_str_len(key);

  // Empty string is represented by the root
  if (key_len == 0) {
    if (trie->root->data == NULL) {
      trie->root->data = data;
      trie->size++;

      return lsm_error_ok;
    } else {
      return lsm_error_already_present;
    }
  }

  uint64_t index = 0;
  lsm_trie_node *node = trie->root;
  lsm_trie_node *next_node;

  while (index < key_len) {
    char c = lsm_str_char(key, index);
    lsm_error res = lsm_bt_search((void **)&next_node, &node->bt, c);

    // No child is present yet for this character, so we can insert the string
    // here
    if (res == lsm_error_not_found) {
      lsm_trie_node *new_node;
      LSM_RES(lsm_trie_node_init(&new_node));

      new_node->data = data;

      lsm_str_substr(&new_node->skip, key, index + 1, key_len);
      res = lsm_bt_insert(&node->bt, c, new_node);

      if (res != lsm_error_ok) {
        lsm_trie_node_free(new_node);

        return res;
      }

      trie->size++;

      return lsm_error_ok;
    }

    index++;

    // We compare the remaining part of the key with the node's skip. If cmp is
    // less than the length of the skip, we know they differ and the edge should
    // be split.
    uint64_t cmp = lsm_str_cmp(key, index, &next_node->skip, 0);

    if (cmp < lsm_str_len(&next_node->skip)) {
      lsm_trie_node *split_node;
      LSM_RES(lsm_trie_node_init(&split_node));

      // split_node replaces the original node as the new child node
      // bottom_node here is always the same value as next_node
      lsm_trie_node *bottom_node;
      lsm_bt_replace((void **)&bottom_node, &node->bt, c, split_node);

      // The old next node now becomes the child of split_node
      lsm_bt_insert(&split_node->bt, lsm_str_char(&next_node->skip, cmp),
                    next_node);

      // split_node's skip has not been initialized yet, so we can simply
      // overwrite it with bottom_node's skip
      split_node->skip = bottom_node->skip;

      // The new node splits the edge into two parts, so the new split node will
      // have the first part of the skip (minus the one character) as its
      // skip
      lsm_str_substr(&next_node->skip, &split_node->skip, cmp + 1,
                     lsm_str_len(&split_node->skip));

      // The old node keeps the first part of the skip
      lsm_str_truncate(&split_node->skip, cmp);

      next_node = split_node;
    }

    node = next_node;
    index += cmp;
  }

  // This catches the edge case where the exact node for the string is already
  // present in the trie
  if (node->data != NULL) {
    return lsm_error_already_present;
  }

  node->data = data;
  trie->size++;

  return lsm_error_ok;
}

lsm_error lsm_trie_search(void **out, const lsm_trie *trie,
                          const lsm_str *key) {
  uint64_t key_len = lsm_str_len(key);

  if (key_len == 0) {
    if (trie->root->data != NULL) {
      if (out != NULL) {
        *out = trie->root->data;
      }

      return lsm_error_ok;
    } else {
      return lsm_error_not_found;
    }
  }

  uint64_t index = 0;
  lsm_trie_node *node = trie->root;
  lsm_trie_node *next_node;

  while (index < key_len) {
    char c = lsm_str_char(key, index);
    lsm_error res = lsm_bt_search((void **)&next_node, &node->bt, c);

    if (res != lsm_error_ok) {
      return res;
    }

    index++;

    uint64_t cmp = lsm_str_cmp(key, index, &next_node->skip, 0);

    // If we end in the middle of an edge, we definitely haven't found the node
    if (cmp != lsm_str_len(&next_node->skip)) {
      return lsm_error_not_found;
    }

    node = next_node;
    index += cmp;
  }

  if (node->data == NULL) {
    return lsm_error_not_found;
  }

  if (out != NULL) {
    *out = node->data;
  }

  return lsm_error_ok;
}

lsm_error lsm_trie_remove(void **out, lsm_trie *trie, const lsm_str *key) {
  uint64_t key_len = lsm_str_len(key);

  if (key_len == 0) {
    if (trie->root->data != NULL) {
      if (out != NULL) {
        *out = trie->root->data;
      }

      trie->root->data = NULL;
      trie->size--;

      return lsm_error_ok;
    } else {
      return lsm_error_not_found;
    }
  }

  uint64_t index = 0;
  lsm_trie_node *parent = trie->root;
  lsm_trie_node *child;

  while (index < key_len) {
    char c = lsm_str_char(key, index);
    lsm_error res = lsm_bt_search((void **)&child, &parent->bt, c);

    if (res != lsm_error_ok) {
      return res;
    }

    index++;

    uint64_t cmp = lsm_str_cmp(key, index, &child->skip, 0);

    // If we end in the middle of an edge, we definitely haven't found the node
    if (cmp != lsm_str_len(&child->skip)) {
      return lsm_error_not_found;
    }

    index += cmp;

    // This context is needed for the removal
    if (index < key_len) {
      parent = child;
    }
  }

  if (child->data == NULL) {
    return lsm_error_not_found;
  }

  if (out != NULL) {
    *out = child->data;
  }

  child->data = NULL;

  trie->size--;

  return lsm_error_ok;
}
