#include <stdio.h>
#include <string.h>

#include "lnm/http/consts.h"
#include "lnm/http/loop.h"
#include "lnm/log.h"
#include "lnm/loop.h"
#include "lsm/store.h"

#include "lander.h"
#include "ltm/template.h"

static const char index_page[] =
    "<!DOCTYPE html>\n"
    "<html>\n"
    "  <body>\n"
    "    <h1>r8r.be</h1>\n"
    "    <p>This is the URL shortener and pastebin accompanying my site, <a "
    "href=\"https://rustybever.be\">The Rusty Bever</a>.</p>\n"
    "  </body>\n"
    "</html>\n";

lnm_http_step_err lander_get_root(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;

  lnm_http_res_body_set_buf(&ctx->res, (char *)index_page,
                            sizeof(index_page) - 1, false);
  lnm_http_res_add_header(&ctx->res, lnm_http_header_content_type, "text/html",
                          false);

  return lnm_http_step_err_done;
}

lnm_http_step_err lander_get_redirect(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;

  // For redirects, the URL is stored as an in-memory attribute
  lsm_str *url_attr_val;

  // This shouldn't be able to happen
  if (lsm_entry_attr_get(&url_attr_val, c_ctx->entry, lander_attr_type_url) !=
      lsm_error_ok) {
    lnm_lerror("lander", "%s",
               "Entry of type redirect detected without URL attribute");

    ctx->res.status = lnm_http_status_internal_server_error;
    lsm_entry_close(c_ctx->entry);
    c_ctx->entry = NULL;

    return lnm_http_step_err_res;
  }

  lnm_http_res_add_header_len(&ctx->res, lnm_http_header_location,
                              (char *)lsm_str_ptr(url_attr_val),
                              lsm_str_len(url_attr_val), false);

  ctx->res.status = lnm_http_status_moved_permanently;

  return lnm_http_step_err_done;
}

lnm_err lander_entry_data_streamer(uint64_t *written, char *buf,
                                   lnm_http_conn *conn, uint64_t offset,
                                   uint64_t len) {
  // TODO respect offset variable

  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;

  lsm_entry_data_read(written, buf, c_ctx->entry, len);

  return lnm_err_ok;
}

lnm_err lander_template_streamer(size_t *written, char *buf,
                                 lnm_http_conn *conn, uint64_t offset,
                                 uint64_t len) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;

  // TODO respect offset variable

  ltm_instance_write(written, buf, len, c_ctx->instance);

  return lnm_err_ok;
}

ltm_err lander_data_to_template(size_t *written, char *buf, size_t len,
                                void *data) {
  lsm_entry_handle *entry = data;

  lsm_entry_data_read(written, buf, entry, len);

  return ltm_err_ok;
}

lnm_http_step_err lander_get_paste(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_gctx *c_gctx = ctx->g->c;
  lander_ctx *c_ctx = ctx->c;

  ltm_template_instantiate(&c_ctx->instance, c_gctx->templates.paste);
  /* ltm_instance_block_add_var(c_ctx->instance, ltm_instance_block_type_buf,
   * lsm_str_ptr(c_ctx->entry)) */
  ltm_instance_block_add_var_fn(c_ctx->instance, "paste",
                                lander_data_to_template, c_ctx->entry,
                                lsm_entry_data_len(c_ctx->entry));

  lnm_http_res_body_set_fn(&ctx->res, lander_template_streamer,
                           ltm_instance_size(c_ctx->instance));
  lnm_http_res_add_header(&ctx->res, lnm_http_header_content_type, "text/html",
                          false);

  return lnm_http_step_err_done;
}

lnm_http_step_err lander_get_file(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;

  lnm_http_res_body_set_fn(&ctx->res, lander_entry_data_streamer,
                           lsm_entry_data_len(c_ctx->entry));
  lander_attr_to_header(ctx, lander_attr_type_content_type,
                        lnm_http_header_content_type);

  lsm_str *value;
  char *buf;

  if (lsm_entry_attr_get(&value, c_ctx->entry, lander_attr_type_file_name) ==
      lsm_error_ok) {
    buf = malloc(24 + lsm_str_len(value));
    int len = lsm_str_len(value);
    sprintf(buf, "attachment; filename=\"%.*s\"", len, lsm_str_ptr(value));
  } else {
    buf = malloc(11);
    strcpy(buf, "attachment");
  }

  lnm_http_res_add_header(&ctx->res, lnm_http_header_content_disposition, buf,
                          true);

  return lnm_http_step_err_done;
}

lnm_http_step_err lander_get_index(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_gctx *c_gctx = ctx->g->c;
  lander_ctx *c_ctx = ctx->c;

  ltm_template_instantiate(&c_ctx->instance, c_gctx->templates.index);

  const lsm_str *key;
  lsm_entry_key(&key, c_ctx->entry);

  ltm_instance_block_add_var(c_ctx->instance, "title_key",
                             ltm_instance_block_type_buf,
                             (void *)lsm_str_ptr(key), lsm_str_len(key));

  lsm_str *prefix;
  lsm_str_init_copy_n(&prefix, lsm_str_ptr(key), lsm_str_len(key));
  lsm_str_append_c(prefix, "_", 1);

  uint64_t prefix_len = lsm_str_len(prefix);

  lsm_store_iterator *iter;
  lsm_store_iter(&iter, c_gctx->store, prefix);

  lsm_entry_handle *handle;
  while (lsm_store_iter_next_read(&handle, iter) == lsm_error_ok) {
    const lsm_str *entry_key;
    lsm_entry_key(&entry_key, handle);

    lander_entry_type t;
    lsm_entry_attr_get_uint8_t((uint8_t *)&t, handle,
                               lander_attr_type_entry_type);

    ltm_instance *nested;

    switch (t) {
    case lander_entry_type_redirect:
      ltm_instance_block_add_nested(&nested, c_ctx->instance, "redirect");

      ltm_instance_block_add_var(
          nested, "full_key", ltm_instance_block_type_buf,
          (void *)lsm_str_ptr(entry_key), lsm_str_len(entry_key));
      ltm_instance_block_add_var(nested, "short_key",
                                 ltm_instance_block_type_buf,
                                 (void *)(lsm_str_ptr(entry_key) + prefix_len),
                                 lsm_str_len(entry_key) - prefix_len);
      break;
    case lander_entry_type_file:
    case lander_entry_type_paste:
      ltm_instance_block_add_nested(&nested, c_ctx->instance,
                                    t == lander_entry_type_file ? "file"
                                                                : "paste");

      ltm_instance_block_add_var(
          nested, "full_key", ltm_instance_block_type_buf,
          (void *)lsm_str_ptr(entry_key), lsm_str_len(entry_key));

      // Add the filename as the link name if known, or just the key otherwise
      lsm_str *filename;
      if (lsm_entry_attr_get(&filename, handle, lander_attr_type_file_name) ==
          lsm_error_ok) {
        ltm_instance_block_add_var(
            nested, "filename", ltm_instance_block_type_buf,
            (void *)lsm_str_ptr(filename), lsm_str_len(filename));
      } else {
        ltm_instance_block_add_var(
            nested, "filename", ltm_instance_block_type_buf,
            (void *)(lsm_str_ptr(entry_key) + prefix_len),
            lsm_str_len(entry_key) - prefix_len);
      }
      break;
    }

    // TODO this should be done after writing the template
    lsm_entry_close(handle);
  }

  lnm_http_res_body_set_fn(&ctx->res, lander_template_streamer,
                           ltm_instance_size(c_ctx->instance));
  lnm_http_res_add_header(&ctx->res, lnm_http_header_content_type, "text/html",
                          false);

  lsm_store_iter_free(iter);
  lsm_str_free(prefix);

  return lnm_http_step_err_done;
}

lnm_http_step_err lander_get_entry(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;
  lnm_http_loop_gctx *gctx = ctx->g;
  lander_gctx *c_gctx = gctx->c;

  const char *key_s = &ctx->req.path.s[ctx->req.path.groups[1].rm_so];
  int key_len = ctx->req.path.groups[1].rm_eo - ctx->req.path.groups[1].rm_so;

  lsm_str *key;
  lsm_str_init_copy_n(&key, (char *)key_s, key_len);
  lsm_error lsm_res = lsm_store_open_read(&c_ctx->entry, c_gctx->store, key);
  lsm_str_free(key);

  switch (lsm_res) {
  case lsm_error_ok:
    break;
  case lsm_error_not_found:
    ctx->res.status = lnm_http_status_not_found;
    return lnm_http_step_err_res;
  default:
    ctx->res.status = lnm_http_status_internal_server_error;
    return lnm_http_step_err_res;
  }

  lander_entry_type t;
  lsm_entry_attr_get_uint8_t((uint8_t *)&t, c_ctx->entry,
                             lander_attr_type_entry_type);

  lnm_http_step_err res;

  switch (t) {
  case lander_entry_type_redirect:
    res = lander_get_redirect(conn);
    break;
  case lander_entry_type_paste:
    res = lander_get_paste(conn);
    break;
  case lander_entry_type_file:
    res = lander_get_file(conn);
    break;
  case lander_entry_type_index:
    res = lander_get_index(conn);
    break;
  }

  return res;
}
